#include "spi_flash.h"

#define NRF_LOG_MODULE_NAME spi_flash
#include "nrf_log.h"
#include "nrf_log_ctrl.h"
NRF_LOG_MODULE_REGISTER();

#include <stdbool.h>
#include <stdint.h>
#include "nrf.h"
#include "nrf_gpio.h"
#include "app_error.h"
#include "sdk_config.h"
#include "boards.h"


// MX25R6435F command opcodes
#define SF25_OP_READ_DATA       0x03
#define SF25_OP_WRITE_DATA      0x02
#define SF25_OP_READ_STAT       0x05
#define SF25_OP_WRITE_ENABLE    0x06
#define SF25_OP_WRITE_DISABLE   0x04
#define SF25_OP_ERASE_SECTOR    0x20
#define SF25_OP_ERASE_BLOCK     0x52
#define SF25_OP_READ_ID         0x9F
#define SF25_OP_NOP             0x00



#define SF25_CSN_PIN            NRF_GPIO_PIN_MAP(0, 22)
#define SF25_SCK_PIN            NRF_GPIO_PIN_MAP(0, 20)
#define SF25_MOSI_PIN           NRF_GPIO_PIN_MAP(0, 23)
#define SF25_MISO_PIN           NRF_GPIO_PIN_MAP(0, 13)




//#define HIGH_SCK nrf_gpio_pin_set(SF25_SCK_PIN)
//#define LOW_SCK  nrf_gpio_pin_clear(SF25_SCK_PIN);

static bool m_is_init = false;


void spi_flash_init(void)
{
    if (m_is_init) return;

    // pull high CS pin
    nrf_gpio_pin_set(SF25_CSN_PIN);
    nrf_gpio_cfg_output(SF25_CSN_PIN);

    // initialize SPI
    nrf_gpio_pin_clear(SF25_SCK_PIN);
    nrf_gpio_cfg_output(SF25_SCK_PIN);
    nrf_gpio_pin_clear(SF25_MOSI_PIN);
    nrf_gpio_cfg_output(SF25_MOSI_PIN);
    nrf_gpio_cfg_input(SF25_MISO_PIN, NRF_GPIO_PIN_PULLDOWN);

    m_is_init = true;
}

void spi_flash_uninit(void)
{
    if (!m_is_init) return;

    m_is_init = false;

    nrf_gpio_cfg_default(SF25_SCK_PIN);
    nrf_gpio_cfg_default(SF25_MOSI_PIN);
    nrf_gpio_cfg_default(SF25_MISO_PIN);
    nrf_gpio_cfg_default(SF25_CSN_PIN);
}

static void spi_flash_write_bytes(const uint8_t *p_data, uint8_t len)
{
    for (uint8_t n = 0; n < len; n++)
    {
        uint8_t val = *(p_data + n);

        // ORDER = MsbFirst
        for (uint8_t nb = 0x80; nb > 0; nb >>= 1)
        {
            if (val & nb)
                nrf_gpio_pin_set(SF25_MOSI_PIN);
            else
                nrf_gpio_pin_clear(SF25_MOSI_PIN);

            __NOP();
            __NOP();

            // SPI mode 0, CPHA = Leading, CPOL = ActiveHigh
            nrf_gpio_pin_set(SF25_SCK_PIN);
            __NOP();
            __NOP();
            nrf_gpio_pin_clear(SF25_SCK_PIN);
        }
    }
}

static void spi_flash_read_bytes(uint8_t *p_data, uint32_t len)
{
    for (uint32_t n = 0; n < len; n++)
    {
        uint8_t val = 0;

        // ORDER = MsbFirst
        for (uint8_t nb = 0x80; nb > 0; nb >>= 1)
        {
            __NOP();
            __NOP();

            // SPI mode 0, CPHA = Leading, CPOL = ActiveHigh
            nrf_gpio_pin_set(SF25_SCK_PIN);
            val = (val << 1) | nrf_gpio_pin_read(SF25_MISO_PIN);
//          __NOP();
//          __NOP();
            nrf_gpio_pin_clear(SF25_SCK_PIN);
        }

        *(p_data + n) = val;
    }
}

static ret_code_t spi_flash_start_op_4(uint8_t opcode, uint32_t addr)
{
    if (addr >= FLASH_SIZE) return NRF_ERROR_INVALID_PARAM;

    uint8_t buf[4];
    buf[0] = opcode;
    buf[1] = (uint8_t)((addr >> 16) & 0xFF);
    buf[2] = (uint8_t)((addr >>  8) & 0xFF);
    buf[3] = (uint8_t)((addr >>  0) & 0xFF);

    nrf_gpio_pin_clear(SF25_CSN_PIN);
    spi_flash_write_bytes(buf, sizeof(buf));

    return NRF_SUCCESS;
}

ret_code_t spi_flash_read_data(uint32_t addr)
{
    if (!m_is_init) return NRF_ERROR_INVALID_STATE;

    return spi_flash_start_op_4(SF25_OP_READ_DATA, addr);
}

ret_code_t spi_flash_data_get(void *buff, uint32_t len)
{
    if (!m_is_init) return NRF_ERROR_INVALID_STATE;

    // read SPI flash data
    nrf_gpio_pin_clear(SF25_CSN_PIN);
    spi_flash_read_bytes(buff, len);

    // pull high CS pin
    nrf_gpio_pin_set(SF25_CSN_PIN);

    return NRF_SUCCESS;
}
